/* Define to type of ssize_t, if not on this platform.  */
#undef ssize_t

/* Define if you have the <sys/select.h> include file */
#undef HAVE_SYS_SELECT_H

/* Define if you don't have fd_set */
#undef NO_FD_SET

/* Define if you don't have sys_errlist */
#undef NO_SYS_ERRLIST

/* Define if want to compile with ASSERT() statements */
#undef WITH_ASSERT

/* Define if want to compile with additional debugging code */
#undef WITH_DEBUG

/* Define if want to compile with hooks for testing */
#undef WITH_TEST

/* Define if sockaddr_un in <sys/un.h> contains the sun_len component */
#undef HAVE_SOCKADDR_UN_SUN_LEN

/* Define if we have f{set,get}pos functions */
#undef HAVE_FPOS

/* Define if we need cross-process locking */
#undef USE_LOCKING

/* Define if va_arg(arg, long double) crashes the compiler. */
#undef HAVE_VA_ARG_LONG_DOUBLE_BUG

/* Don't know what this stuff is for */
#undef HAVE_MATHLIB
#undef WITH_DOMESTIC
#undef WITH_EXPORT
#undef WITH_GLOBAL
